#include "ExampleFrameListener.h"

#include <iostream>
#include <cassert>

using namespace std;

#include <btBulletCollisionCommon.h>
#include <Ogre.h>
#include <OIS/OIS.h>

using namespace Ogre;

#include "Player.h"

#define PRINT_BTVECTOR3(btvec) \
    cout << btvec[0] << " " << btvec[1] << " " << btvec[2] << endl;

void registerTriMeshes(SceneManager* sceneMgr, btCollisionWorld* colWorld)
{
    SceneManager::MovableObjectIterator 
        i = sceneMgr->getMovableObjectIterator("Entity");
        
    while (i.hasMoreElements()) {
        Entity* entity = static_cast<Entity*>(i.getNext());
        cout << "Name: " << entity->getName() << endl;
        
        btTriangleMesh* btMesh = new btTriangleMesh();
        
        MeshPtr mesh = entity->getMesh();
        Mesh::SubMeshIterator j = mesh->getSubMeshIterator();
        while (j.hasMoreElements()) {
            SubMesh* submesh = j.getNext(); 
            
            int idxStart = submesh->indexData->indexStart;
            int nIdx = submesh->indexData->indexCount;
            
            HardwareIndexBuffer* idxBuffer 
                = submesh->indexData->indexBuffer.get();
                
            HardwareVertexBufferSharedPtr virtBuffer;
            VertexDeclaration* virtDecl;
            
            if (submesh->useSharedVertices) {
                virtDecl = mesh->sharedVertexData->vertexDeclaration;
                assert(mesh->sharedVertexData->vertexBufferBinding->getBufferCount() > 0);
                virtBuffer = mesh->sharedVertexData->vertexBufferBinding->getBuffer(0);
            } else {
                virtDecl = submesh->vertexData->vertexDeclaration;
                assert(submesh->vertexData->vertexBufferBinding->getBufferCount() > 0);
                virtBuffer = submesh->vertexData->vertexBufferBinding->getBuffer(0);
            }
            
            unsigned char* pVert = static_cast<unsigned char*>(virtBuffer->lock(HardwareBuffer::HBL_READ_ONLY));

            // need to lock the buffer since vertex data resides on GPU
            // and we need synchronization
            unsigned short* sindices = NULL;
            unsigned long* lindices = NULL;
            
            if (idxBuffer->getType() == HardwareIndexBuffer::IT_16BIT) {
                sindices = static_cast<unsigned short*>(idxBuffer->lock(HardwareBuffer::HBL_READ_ONLY));
            } else if (idxBuffer->getType() == HardwareIndexBuffer::IT_32BIT) {
                lindices = static_cast<unsigned long*>(idxBuffer->lock(HardwareBuffer::HBL_READ_ONLY));
            } else {
                assert(true == false);
            }
            
            const VertexElement* elm = virtDecl->findElementBySemantic(VES_POSITION, 0);
            int offset = elm->getOffset();
            assert(elm->getType() == VET_FLOAT3);
            
            for (int k = idxStart; k < idxStart + nIdx; k += 3) {
                unsigned int indices[3];
                btVector3 vertices[3];
                
                if (idxBuffer->getType() == HardwareIndexBuffer::IT_16BIT) {
                    for (int l = 0; l < 3; ++l) {
                        indices[l] = sindices[k + l];
                    }
                } else {
                    for (int l = 0; l < 3; ++l) {
                        indices[l] = lindices[k + l];
                    }
                }
                
                for (int l = 0; l < 3; ++l) { // for each vertex
                    //cout << indices[l] << endl;
                    Real* posVert = (Real*)(pVert + indices[l] * virtBuffer->getVertexSize() + offset);
                    for (int m = 0; m < 3; ++m) { // for each vertex component
                        vertices[l][m] = posVert[m];
                    }
                    //cout << vertices[l][0] << " " << vertices[l][1] << " " << vertices[l][2] << endl;
                }
                
                btMesh->addTriangle(vertices[0], vertices[1], vertices[2]);
            }
                
            idxBuffer->unlock();
            virtBuffer->unlock();
        }
        
        // uses a bounding volume hierarchy to do medium phase culling
        btBvhTriangleMeshShape* btMeshShape = new btBvhTriangleMeshShape(btMesh, true, true);
        
        btCollisionObject* btObj = new btCollisionObject();
        btObj->setCollisionShape(btMeshShape);
        
        Vector3 pos = entity->getParentSceneNode()->getPosition();
        Quaternion orient = entity->getParentSceneNode()->getOrientation();
        Vector3 scale = entity->getParentSceneNode()->getScale();
        
        //cout << "scale: " << scale[0] << " " << scale[1] << " " << scale[2] <<  endl;
        btMeshShape->setLocalScaling(btVector3(scale[0], scale[1], scale[2]));
        btMeshShape->setMargin(0.0);
        
        //cout << "position: " << pos << endl;
        
        btTransform btTrans;
        
        btTrans.setIdentity();
        btTrans.setOrigin(btVector3(pos[0], pos[1], pos[2]));
        btTrans.setRotation(btQuaternion(orient[1], orient[2], orient[3], orient[0]));
        btObj->setWorldTransform(btTrans);
        
        // the last 2 parameters are bit strings representing group membership
        // and the groups which it is allowed to collide with
        colWorld->addCollisionObject(btObj, 2, 1);
    }
}

/*
 * Extends the ExampleFrameListener for collision detection duties.
 */
class MyFrameListener : public ExampleFrameListener
{
protected:
    SceneManager* mSceneMgr;
    
    btDefaultCollisionConfiguration* mColConfig;
    btCollisionWorld* mColWorld;
    btCollisionDispatcher* mDispatcher;
    btAxisSweep3* mBroadphase;
    
    Player* player;
    
public:
    MyFrameListener(
        RenderWindow* win, Camera* cam, SceneManager* sceneMgr,
        bool bufferedKeys = false, bool bufferedMouse = false
        ) 
        : ExampleFrameListener(win, cam, bufferedKeys, bufferedMouse), mSceneMgr(sceneMgr) 
    {
        //
        // initialize bullets collision detection system
        //
        btVector3 worldAabbMin(-1000,-1000,-1000);
        btVector3 worldAabbMax(1000,1000,1000);
        
        mBroadphase = new btAxisSweep3(worldAabbMin, worldAabbMax);
        mColConfig = new btDefaultCollisionConfiguration();
        mDispatcher = new btCollisionDispatcher(mColConfig);
        mColWorld = new btCollisionWorld(mDispatcher, mBroadphase, mColConfig);
    
        // registers all of the meshes currently in the scene manager
        // as trimesh colliders in the bullet btCollisionWorld object
        // Each mesh will be assigned to collision group 2. Our
        // character should be assigned to collision group 1.
        registerTriMeshes(mSceneMgr, mColWorld);
        
        player = new Player(mSceneMgr, mColWorld);
        player->init();
        player->move(4.0, 0.0, 0.0);
    }

    virtual ~MyFrameListener() 
    {
        delete mColWorld;
        delete mDispatcher;
        delete mColConfig;
        delete mBroadphase;
    }

    bool frameStarted(const FrameEvent& evt)
    {
        // do whatever the example frame listener did for frame started
        // processUnbufferedKeyInput is called in here
        if (!ExampleFrameListener::frameStarted(evt)) return false;
        
        // Here's where all the collision detection magic occurs. 
        // We need to call this before performing our character cylinder 
        // world check. We do not need to call this function if we wish to simply
        // perform ray queries.
        mColWorld->performDiscreteCollisionDetection();
        player->updateFrameBegin(evt);
        
        return true;
    }
    
    bool frameEnded(const FrameEvent& evt)
    {
        if (!ExampleFrameListener::frameEnded(evt)) return false;
        
        player->updateFrameEnd(evt);

        return true;
    }
    
    bool processUnbufferedKeyInput(const FrameEvent& evt)
    {
        if (!ExampleFrameListener::processUnbufferedKeyInput(evt)) return false;
        
        if (mKeyboard->isKeyDown(OIS::KC_I)) {
            player->move(Player::FORWARD);
        }
        
        if (mKeyboard->isKeyDown(OIS::KC_K)) {
            player->move(Player::BACKWARD);
        } 
        
        if (mKeyboard->isKeyDown(OIS::KC_J)) {
            player->move(Player::TURN_LEFT);
        }
        if (mKeyboard->isKeyDown(OIS::KC_L)) {
            player->move(Player::TURN_RIGHT);
        }
        
        return true;
    }
};
